class LocalClient

  def initialize(share, logger, file_logger = nil)
    @share = share
    @logger = logger
    @file_logger = file_logger
    @changed_directoryies_mutex = Mutex.new
    @changed_directoryies = []
    @fail_count = 0
  end

  def auth(protonet_user, protonet_password)
    with_network_errors_caught do
      @share.remote.auth!(protonet_user, protonet_password)
    end
  end

  def directory_changed(directory)
    @changed_directoryies_mutex.synchronize do
      @changed_directoryies << directory
    end
  end

  def needs_scan?
    @changed_directoryies_mutex.synchronize do
      @changed_directoryies.any?
    end
  end

  def scanned!
    @changed_directoryies_mutex.synchronize do
      @changed_directoryies = []
    end
  end

  def box_reachable?
    @share.remote.reachable?
  end

  def start_local_scan
    fsevent = FSEvent.new

    fsevent.watch @share.path do |directories|
      directories.each do |dir|
        directory_changed(dir)
      end
    end

    @local_scan_thread = Thread.new { fsevent.run }
  end

  def with_network_errors_caught
    yield
    @fail_count = 0
  rescue Exception => e
    @logger.call("box_unreachable")
    @file_logger.call(e) # Log Exception only to file.

    if box_reachable?
      @fail_count += 1
      raise if @fail_count > 10
    end

    sleep 10
  end

  def run
    ppid = Process.ppid

    start_local_scan

    @share.status_monitor = @share.remote.status_monitor = @logger

    @share.scan

    while (true)
      if needs_scan?
        scanned!
        @share.scan
      end

      with_network_errors_caught do
        begin
          @share.remote.push
          @share.remote.pull
          @share.apply
          @logger.call("no_changes_detected")
          sleep 10
        rescue Errno::ENOENT => e
          @logger.call("error", e.inspect)
          @logger.call("rescan_needed")
        end
      end

      Process.kill("CONT", ppid)
    end

    @local_scan_thread.join
  end
end