require 'rrdiff'

class LocalFile
  attr_reader :path

  FS_UTF8_NORMALISATION=
  case RUBY_PLATFORM
  when /linux/
    :c
  when /darwin/
    :d
  else
    raise "LocalFile unsupported platform #{RUBY_PLATFORM}"
  end

  class << self
    def in_jail?(base, path)
      base_expanded = File.expand_path(base + "/")
      full_expanded = File.expand_path("#{base}/#{path}")
      File.expand_path(full_expanded[0..base_expanded.size]) == base_expanded
    end
  end

  def initialize(path, force_existance = false)
    @path = ActiveSupport::Multibyte::Unicode.normalize(path, FS_UTF8_NORMALISATION)
  end

  def to_s
    "LocalFile(path:#{@path})"
  end

  def up_to_date?(remote_sha1, remote_mtime, remote_mode)
    # FIXME: apply mode and mtime directly?
    stat = File.stat(@path)
    (remote_sha1 == sha1) && (stat.mode == remote_mode) && (stat.mtime.to_i == remote_mtime)
  end

  def exists?
    File.exists?(@path)
  end

  def create_from_remote(remote, remote_mtime, remote_mode)
    write(remote.retrieve(@path), remote_mtime, remote_mode)
  end

  def update_from_remote(remote, remote_mtime, remote_mode)
    patch(remote.retrieve_patch(@path, signature), remote_mtime, remote_mode)
    ensure_mode(remote_mode)
    ensure_mtime(remote_mtime)
  end

  def ensure_mtime(mtime)
    ts = Time.at(mtime)
    File.utime(ts, ts, @path)
  end

  def ensure_mode(mode)
    File.chmod(mode, @path)
  end

  def patch(patch, mtime, mode)
    raise Errno::ENOENT unless exists?

    patch_file = tempfile
    patch_file.write(patch)
    patch_file.close

    new_file = tempfile
    new_file.close

    RRDiff.patch(@path, patch_file.path, new_file.path)
    # FIXME: needed because we could have a "-r--r--r--" File
    FileUtils.rm_f(@path)
    FileUtils.cp(new_file.path, @path)
    ensure_mode(mode)
    ensure_mtime(mtime)
  ensure
    patch_file.unlink
    new_file.unlink
  end

  def delta(other_signature)
    raise Errno::ENOENT unless exists?
    
    sig_file = tempfile
    sig_file.write(other_signature)
    sig_file.close

    delta_file = tempfile
    delta_file.close

    RRDiff.delta(@path, sig_file.path, delta_file.path)

    read(delta_file.path)
  ensure
    sig_file.unlink
    delta_file.unlink
  end

  def signature
    return nil unless exists?

    sfile = tempfile
    RRDiff.signature(@path, sfile.path)
    result = sfile.read.force_encoding("BINARY")
    sfile.close(true)
    result
  end

  def read(path = @path)
    IO.read(path).force_encoding("BINARY")
  end

  def rmdir
    FileUtils.rmdir(@path) if File.directory?(@path)
  end

  def mkdir 
    FileUtils.mkdir_p(@path) unless File.directory?(@path)
  end

  def rm
    File.delete(@path) if File.exists?(@path)
  end

  def write(content, mtime, mode)
    dir_made = false
    IO.write(@path, content)
    ensure_mode(mode)
    ensure_mtime(mtime)
  rescue Errno::ENOENT => e
    raise if dir_made
    dir = File.dirname(@path)
    FileUtils.mkdir_p(dir)
    dir_made = true
    retry
  end

  private 
    def sha1
      Digest::SHA1.file(@path).hexdigest rescue nil
    end

    def tempfile
      tmp = Tempfile.new("proto")
      tmp.binmode
      tmp
    end
end