require 'rest-client'
require 'macaddr'

class Remote < ActiveRecord::Base
  VERSION=20131217
  MAX_MYSQL_TRANSACTION_NUMBER_AS_LONG_AS_ITS_32BIT=2**31-1

  cattr_accessor :sync_uri
  
  class NoConnectionError < RuntimeError; end
  
  belongs_to :share

  has_many :commands, -> { order(:remote_txn) }

  attr_accessor :status_monitor

  def reset
    commands.destroy_all
    update_attributes(remote_txn:0, local_txn:0)
  end

  def auth!(user, password)
    response = JSON.parse(post("auth", user:user, password:password, device: Mac.addr, version:VERSION))
    update_attribute(:token, response["token"])
  end

  def reachable?
    [] == JSON.parse(get("changes", txn:MAX_MYSQL_TRANSACTION_NUMBER_AS_LONG_AS_ITS_32BIT))
  rescue Exception => e
    false
  end

  def pull
    transaction do
      new_commands.each do |command|
        commands << Command.new(:command => command)
      end
      new_remote_txn = [commands.maximum(:remote_txn) || 0, remote_txn].max
      self.update_attributes(remote_txn:new_remote_txn)
    end
  end

  def apply_remote(file)
    if file.deleted?
      if file.is_directory?
        rmdir(file.path)
      else
        rm(file.path)
      end
    else
      if file.is_directory?
        mkdir(file.path)
      else
        local_file = LocalFile.new(file.path)

        if remote_signature = signature(file.path)
          patch(file.path, local_file.delta(remote_signature), file.mtime, file.mode)
        else
          put(file.path, local_file.read, file.mtime, file.mode)
        end
      end
    end
  end

  def push
    Dir.chdir(share.path) do 
      changes = share.sync_files.where("txn > ?", local_txn).order("txn asc").to_a

      return unless changes.size > 0

      # puts "send_changes: changes:#{changes.size}, local_txn:#{local_txn}"

      cnt = changes.size
      inx = 1
      changes.each do |file|
        log_status("uploading", inx, cnt)
        inx += 1
        apply_remote(file)
        update_attribute(:local_txn, file.txn)
      end      
    end
  end

  def rm(path)
    post "rm", path:path
  end

  def rmdir(path)
    post "rmdir", path:path
  end

  def mkdir(path)
    post "mkdir", path:path
  end

  def put(path, content, mtime, mode)
    post "put", path:path, content:content, mtime:mtime, mode:mode
  end

  def patch(path, patch, mtime, mode)
    post "patch", path:path, patch:patch, mtime:mtime, mode:mode
  end

  def signature(path)
    post "signature", path:path
  end

  def retrieve_patch(path, local_signature)
    post "retrieve_patch", path:path, local_signature:local_signature
  end

  def retrieve(path)
    post "retrieve", path:path
  end

  def new_commands
    JSON.parse(get("changes", txn:remote_txn))
  end

  def get path, params
    $journal.info("R:#{path}") if $journal

    RestClient.get "#{Remote.sync_uri}/#{path}", :params => params.merge('token' => token, 'version' => VERSION)
  end

  def post path, params
    $journal.info("R:#{path} #{params[:path]}") if $journal

    RestClient.post("#{Remote.sync_uri}/#{path}", params.merge(multipart: true, 'token' => token, 'version' => VERSION)).force_encoding('BINARY')
  rescue RestClient::ResourceNotFound
    nil
  end

  private 
    def log_status(message, *args)
      status_monitor.call(message, *args) if status_monitor
    end
end
