require "digest/sha1"

class SyncFile < ActiveRecord::Base
  belongs_to :share
  
  scope :not_seen, -> { where(seen: false) }
  scope :not_deleted, -> { where(deleted: false) }

  validates :share, presence: true

  before_save :update_sha1
  before_save :set_txn_when_changed

  class << self

    def create_or_update_from_path(path, share)
      sync_file = where(path: path, share:share).take || new(path: path, share:share)
      sync_file.update_from_filesystem!
    end

    def ignored?(path)
      return true if path.match(/^\.\/\.protonet/) 
      return true if path.match(/[:\\]/) 
      # our mysql has no real utf-8 support (http://mathiasbynens.be/notes/mysql-utf8mb4)
      return true if path.codepoints.detect { |c| c > 0x010000 }
      false
    end

    def reset_seen
      update_all(seen: false)
    end
  end

  def mark_as_deleted
    self.inode = nil
    self.mode = nil
    self.mtime = nil
    self.size = nil
    self.sha1 = ''
    self.deleted = true
  end

  def stat=(stat)
    if is_directory?
      self.size = nil
      self.mtime = nil
    else
      self.size = stat.size
      self.mtime = stat.mtime.to_i
    end
    # self.inode = stat.ino
    self.mode = stat.mode
    self.deleted = false
  end

  def seen
    self.seen = true
    self.deleted = false
  end

  def to_command
    details = if is_directory?
      if deleted?
        {command:'rmdir'}
      else
        {command:'mkdir', mode:mode, mtime:mtime}
      end
    else
      if deleted?
        {command:'rm'}
      else
        {command:'ensure', mode:mode, mtime:mtime, sha1:sha1}
      end
    end
    {path:path, txn:txn}.merge details
  end

  def update_from_filesystem!
    if stat = (File.stat(path) rescue nil)
      self.is_directory = stat.directory?
      self.stat = stat
    else
      self.mark_as_deleted
    end

    without_txn do 
      save!
    end
  end

  private
    def without_txn
      @skip_txn = true
      yield
    ensure
      @skip_txn = false
    end

    def set_txn_when_changed
      return true if @skip_txn || ! changed?
      # puts "detected change: #{id}:'#{path}' changes:#{changes.keys.inspect}"
      self.txn = share.next_txn
      true
    end

    def update_sha1
      return true if deleted? || is_directory? || ! changed?
      self.sha1 = Digest::SHA1.file(path).hexdigest
      true
    rescue Errno::ENOENT
      puts "Errno::ENOENT in SyncFile.update_sha1 #{path} "
      false
    end
end